#!/usr/bin/env python3

"""
Formatting options to pass to Conan
"""

import typing


def format_options(options: typing.Dict[str, str]) -> typing.List[str]:
    """
    Format the dictionary of options appropriately for Conan 1.x
    """
    options_list = [f"{optkey}={optvalue}" for optkey, optvalue in options.items()]
    return options_list


def format_options_v2(options: typing.Dict[str, str]) -> typing.List[str]:
    """
    Format the dictionary of options appropriately for Conan 2.x
    Expected to be of the form
        -o name/*:option=value
    """
    options_list: typing.List[str] = []
    for optkey, optvalue in options.items():
        assert ":" in optkey
        optkey_split = optkey.split(":")
        options_list.append(f"{optkey_split[0]}/*:{optkey_split[1]}={optvalue}")
    return options_list



# Cruiz workers API

import cruiz.globals as globals

from .common import endmessagethread

if globals.CONAN_MAJOR_VERSION == 1:
    from .v1 import (  # noqa: E402
        arbitrary,
        build,
        cmakebuildtool,
        configinstall,
        create,
        deletecmakecache,
        exportpackage,
        imports,
        install,
        lockcreate,
        meta,
        package,
        packagebinary,
        packagedetails,
        packagerevisions,
        reciperevisions,
        remotesearch,
        removeallpackages,
        removelocks,
        removepackage,
        source,
        testpackage,
    )
else:
    from .v2 import (  # type: ignore[no-redef] # noqa: E402
        arbitrary,
        build,
        configinstall,
        create,
        exportpackage,
        install,
        lockcreate,
        meta,
        packagebinary,
        packagedetails,
        packagerevisions,
        reciperevisions,
        remotesearch,
        removeallpackages,
        removepackage,
        source,
        testpackage,
    )



#!/usr/bin/env python3

"""
Run an arbtitrary Conan command.
It must match one of the Conan verbs though.
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.message import Message, Success

from . import worker


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run an arbitrary command
    """

    # TODO: is it possible to rewrite this? see the v1 as a model
    """
    with worker.ConanWorker(queue, params) as api:
        from conan.cli.command import ConanCommand

        command = Command(api)
        verb_function = getattr(command, params.verb)
        result = command.run(*params.arguments)
        queue.put(Success(result))
    """
    with worker.ConanWorker(queue, params):
        import subprocess

        args = ["conan", params.verb] + params.arguments
        result = subprocess.run(
            args, capture_output=True, encoding="utf-8", errors="ignore"
        )
        queue.put(Success(result.stdout))



#!/usr/bin/env python3

"""
Worker implementation for conan remove
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.message import Message, Success, Stdout

from . import worker


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan remove [-c] ref'
    """
    with worker.ConanWorker(queue, params) as api:
        from conan.api.model import ListPattern

        assert params.package_reference
        ref_pattern = ListPattern(params.package_reference)
        select_bundle = api.list.select(ref_pattern)

        for ref, _ in select_bundle.refs():
            api.remove.recipe(ref)

        queue.put(
            Stdout(
                "Removed packages in the local cache matching pattern "
                f'"{params.package_reference}"'
                " (confirmed)"
                if params.force
                else ""
            )
        )
        queue.put(Success(True))



#!/usr/bin/env python3

"""
Stream data via a multiprocessing queue
"""

from __future__ import annotations

import multiprocessing
import six
import typing

from .colorarma_conversion import convert_from_colorama_to_html

from cruiz.interop.message import Message


class QueuedStreamSix(six.StringIO):
    """
    A stream class that uses multiprocessing.Queue to send messages.
    Complete messages may be passed piecemeal, so that coloured output is generated.
    This buffers up a message, so that it is passed across the process divide as a
    single HTML paragraph.
    """

    def __init__(
        self, queue: multiprocessing.Queue[Message], message_type: typing.Any
    ) -> None:
        super().__init__()
        self._queue = queue
        self._message_type = message_type
        self._block = six.StringIO()

    def write(self, message: str) -> int:
        """
        Write a message.
        """
        result = self._block.write(convert_from_colorama_to_html(message))
        if message.endswith("\n"):
            self._queue.put(self._message_type(self._block.getvalue()))
            self._block = six.StringIO()
        return result



#!/usr/bin/env python3

"""
Get recipe revisions for the specific package reference
"""

from __future__ import annotations

import datetime
import multiprocessing
import typing

from cruiz.interop.reciperevisionsparameters import RecipeRevisionsParameters
from cruiz.interop.message import Message, Success

from . import worker


def invoke(
    queue: multiprocessing.Queue[Message], params: RecipeRevisionsParameters
) -> None:
    """
    Equivalent to

    'conan search -r <remote_name> <reference> -rev'

    RecipeRevisionsParameters has dynamic attributes.
    """
    with worker.ConanWorker(queue, params) as api:
        from conan.internal.conan_app import ConanApp
        from conans.model.recipe_ref import RecipeReference

        remote = api.remotes.get(params.remote_name)
        ref = RecipeReference.loads(params.reference)
        app = ConanApp(api.cache_folder)

        rrevs_and_timestamps: typing.List[typing.Dict[str, str]] = []
        for new_ref in app.remote_manager.get_recipe_revisions_references(ref, remote):
            rrevs_and_timestamps.append(
                {
                    "revision": new_ref.revision,
                    "time": datetime.datetime.utcfromtimestamp(
                        new_ref.timestamp
                    ).strftime("%Y-%m-%dT%H:%M:%S%Z"),
                }
            )

        queue.put(Success(rrevs_and_timestamps))



#!/usr/bin/env python3

"""
Remote browser page
"""

from __future__ import annotations

import html
import pathlib
import shutil
import tarfile
import tempfile
import typing

from qtpy import QtCore, QtGui, QtWidgets

from qtpy import QtWebEngineCore

import cruiz.globals

from cruiz.interop.packagebinaryparameters import PackageBinaryParameters

from cruiz.pyside6.remote_browser_fileview import Ui_remote_browser_fileview

from .page import Page


class _FileNode:
    def __init__(self, path: str, parent: typing.Optional[_FileNode]):
        self.path = path
        self.parent = parent
        self.children: typing.List[_FileNode] = []
        self.is_file = True
        self.link_target: typing.Optional[str] = None
        self.container: typing.Optional[str] = None
        self.is_container = False
        # am hoping this is reusable on the same tarball but outside the context manager
        self.tar_info: typing.Optional[tarfile.TarInfo] = None

    def add_child(self, path: str) -> _FileNode:
        """
        Add a child path
        """
        find = [x for x in self.children if x.path == path]
        if find:
            assert len(find) == 1
            return find[0]
        node = _FileNode(path, self)
        self.children.append(node)
        return node

    def add_tarball(
        self, directory: pathlib.Path, tarball_basename: str, parent: _FileNode
    ) -> None:
        """
        Adding a tarball containing many children
        """
        tar_node = parent.add_child(tarball_basename)
        tar_node.is_container = True
        tarball_path = directory / tarball_basename
        with tarfile.open(tarball_path, "r") as tar:
            for tarinfo in tar:
                path = pathlib.PurePosixPath(tarinfo.name)
                parent = tar_node
                combined_path = pathlib.PurePosixPath()
                for part in path.parts:
                    combined_path /= part
                    parent = parent.add_child(str(combined_path))
                    parent.is_file = False
                    parent.container = tarball_basename
                if tarinfo.isfile():
                    parent.is_file = True  # the leaf is a file
                    parent.tar_info = tarinfo
                else:
                    if tarinfo.issym():
                        parent.is_file = True  # the leaf is a file
                        parent.link_target = tarinfo.linkname  # and also a symlink
                    else:
                        raise AssertionError(
                            f"Unknown entity type for '{tarinfo.name}'"
                        )

    @property
    def child_index(self) -> int:
        """
        Get the child index of this node in relation to its parent
        """
        if self.parent:
            return self.parent.children.index(self)
        return 0


class _PackageBinaryModel(QtCore.QAbstractItemModel):
    def __init__(self, parent: QtCore.QObject) -> None:
        super().__init__(parent)
        self._root: typing.Optional[_FileNode] = None
        self._parent_object = parent

    def set(
        self,
        results: typing.Optional[typing.List[str]],
        folder: typing.Optional[pathlib.Path],
    ) -> None:
        """
        Set the results against the model
        """
        self.beginResetModel()
        if results:
            assert folder
            self._root = _FileNode("/", None)
            self._root.is_file = False
            for file in results:
                if file.endswith(".tgz"):
                    self._root.add_tarball(folder, file, self._root)
                else:
                    self._root.add_child(file)
        else:
            self._root = None
        self.endResetModel()

    def rowCount(self, parent) -> int:  # type: ignore
        if self._root is None:
            return 0
        if parent.isValid():
            node = parent.internalPointer()
            return len(node.children)
        return 1

    def columnCount(self, parent) -> int:  # type: ignore
        # pylint: disable=unused-argument
        if self._root is None:
            return 0
        return 1

    def parent(self, index) -> QtCore.QModelIndex:  # type: ignore
        if not index.isValid():
            return QtCore.QModelIndex()
        node = index.internalPointer()
        if node.parent is None:
            return QtCore.QModelIndex()
        return super().createIndex(node.parent.child_index, 0, node.parent)

    def index(self, row, column, parent):  # type: ignore
        if not parent.isValid():
            return super().createIndex(row, column, self._root)
        parent_node = parent.internalPointer()
        node = parent_node.children[row]
        return super().createIndex(row, column, node)

    def headerData(self, section, orientation, role):  # type: ignore
        if (
            role == QtCore.Qt.DisplayRole
            and orientation == QtCore.Qt.Horizontal
            and section == 0
        ):
            return "Path"
        return None

    def data(self, index, role):  # type: ignore
        node = index.internalPointer()
        if role == QtCore.Qt.DisplayRole and index.column() == 0:
            return pathlib.Path(node.path).name
        if (
            role == QtCore.Qt.ToolTipRole
            and index.column() == 0
            and node.is_file
            and node.link_target
        ):
            return f"Symbolic link to {node.link_target}"
        if role == QtCore.Qt.DecorationRole and index.column() == 0:
            if node.is_file:
                if node.link_target:
                    return self._parent_object.style().standardIcon(
                        QtWidgets.QStyle.SP_FileLinkIcon
                    )
            else:
                return QtWidgets.QFileIconProvider().icon(
                    QtWidgets.QFileIconProvider.Folder
                )
        return None

    def flags(self, index):  # type: ignore
        default_flags = super().flags(index)
        node = index.internalPointer()
        if node.link_target:
            return default_flags & ~QtCore.Qt.ItemIsEnabled
        return default_flags


class _FileViewer(QtWidgets.QDialog):
    def __init__(
        self,
        root: pathlib.Path,
        path: pathlib.Path,
        container: pathlib.Path,
        parent: typing.Optional[QtWidgets.QWidget] = None,
    ) -> None:
        super().__init__(parent)
        self._ui = Ui_remote_browser_fileview()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        self._ui.fileview.page().settings().setAttribute(
            QtWebEngineCore.QWebEngineSettings.LocalContentCanAccessRemoteUrls, True
        )
        if container:
            html_path = _FileViewer._write_html(path, root / container)
            self.setWindowTitle(f"{path} in {container}")
        else:
            html_path = _FileViewer._write_html(root / path, None)
            self.setWindowTitle(str(path))
        self._ui.fileview.load(QtCore.QUrl.fromLocalFile(str(html_path)))

    @staticmethod
    def _write_html(
        path: pathlib.PurePath, archive: typing.Optional[pathlib.Path]
    ) -> pathlib.Path:
        if archive:
            with tarfile.open(archive, "r") as tar:
                tar.extract(str(path), path=archive.parent)
            with (archive.parent / path).open("rt", encoding="utf-8") as data_file:
                contents = data_file.readlines()
            html_path = archive.parent / path
            html_path = html_path.with_suffix(".html")
        else:
            with path.open("rt", encoding="utf-8") as data_file:
                contents = data_file.readlines()
            html_path = pathlib.Path(path.with_suffix(".html"))
        html_path.parent.mkdir(parents=True, exist_ok=True)
        with html_path.open("wt", encoding="utf-8") as html_file:
            url_start = "https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.8.0"
            html_file.write("<html>")
            html_file.write("<head>")
            if cruiz.globals.is_dark_theme():
                html_file.write(
                    f'<link rel="stylesheet" href="{url_start}/styles/dark.min.css">'
                )
            else:
                html_file.write(
                    f'<link rel="stylesheet" href="{url_start}/styles/default.min.css">'
                )
            html_file.write(f'<script src="{url_start}/highlight.min.js"></script>')
            html_file.write(
                '<script src="https://cdnjs.cloudflare.com/ajax/libs/highlightjs-line-numbers.js/2.8.0/highlightjs-line-numbers.min.js"></script>'  # noqa: E501
            )
            html_file.write("<script>hljs.highlightAll();</script>")
            html_file.write("<script>hljs.initLineNumbersOnLoad();</script>")
            html_file.write("</head>")
            html_file.write('<body style="margin: 0; padding: 0">')
            html_file.write("<pre><code>")
            # with line numbers enabled, we lose leading spaces with html.escape,
            # so put them back manually per line
            for line in contents:
                trimmed = line.lstrip()
                if not trimmed:
                    html_file.write(line)  # honour empty lines
                    continue
                escaped = html.escape(trimmed)
                if len(trimmed) == len(line):
                    html_file.write(escaped)
                    continue
                nbsp_prefix = "&nbsp;" * (len(line) - len(trimmed))
                html_file.write(nbsp_prefix + escaped)
            html_file.write("</code></pre>")
            html_file.write("</body>")
            html_file.write("</html>")
        return html_path


class PackageBinaryPage(Page):
    """
    Remote browser page for displaying package binaries
    """

    def setup(self, self_ui: typing.Any) -> None:
        """
        Setup the UI for the page
        """
        self._base_setup(self_ui, 4)
        self._current_pkgref: typing.Optional[str] = None
        self._artifact_folder: typing.Optional[pathlib.Path] = None
        self._model = _PackageBinaryModel(self)
        self._ui.package_binary.setModel(self._model)
        self._ui.package_binary.customContextMenuRequested.connect(self._on_file_menu)

        self._ui.package_binary.doubleClicked.connect(self._on_prev_dclicked)
        self._ui.pbinary_back.clicked.connect(self._on_back)
        self._ui.pbinary_cancel.clicked.connect(self.on_cancel)
        self._ui.pbinary_restart.clicked.connect(self._on_restart)
        self._ui.pbinary_pkgref.customContextMenuRequested.connect(self._on_pkgref_menu)

    def _enable_progress(self, enable: bool) -> None:
        self._ui.pbinary_progress.setMaximum(0 if enable else 1)
        self._ui.pbinary_buttons.setEnabled(not enable)
        self._ui.pbinary_cancel.setEnabled(enable)

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        # pylint: disable=unused-argument
        pkgref = self._previous_pkgref
        self._ui.pbinary_pkgref.setText(pkgref)

        if self._current_pkgref != pkgref:
            self._artifact_folder = pathlib.Path(tempfile.mkdtemp(suffix="cruiz"))
            self._model.set(None, None)
            self._enable_progress(True)
            self._ui.pbinary_groupbox.setEnabled(False)
            params = PackageBinaryParameters(
                reference=pkgref,
                remote_name=self._ui.remote.currentText(),
                where=self._artifact_folder,
            )
            self._context.get_package_details(params, self._complete)

    def _complete(self, results: typing.Any, exception: typing.Any) -> None:
        self._enable_progress(False)
        if exception:
            self._log_details.stderr(str(exception))
            return
        if results is not None:
            self._model.set(results, self._artifact_folder)
            self._ui.package_binary.expandAll()
            self._ui.pbinary_groupbox.setEnabled(True)
            self._current_pkgref = self._previous_pkgref

    def _on_back(self) -> None:
        if self._revisions_enabled:
            self._open_previous_page()
        else:
            # skip package revisions
            self.parent().setCurrentIndex(self.page_index - 2)

    def _on_prev_dclicked(self, index: QtCore.QModelIndex) -> None:
        node = index.internalPointer()
        if not node.is_file:  # type: ignore
            return
        path = node.path  # type: ignore
        container = node.container  # type: ignore
        try:
            assert self._artifact_folder
            _FileViewer(self._artifact_folder, path, container, self).exec_()
        except UnicodeDecodeError:
            QtWidgets.QMessageBox.critical(
                self,
                "Cannot view file contents",
                f"Unable to interpret file {path} in archive {container} " "as text",
            )

    def on_cancel(self) -> None:
        """
        Called when the user cancels the operation.
        """
        self._context.cancel()
        self._enable_progress(False)

    def _on_restart(self) -> None:
        self._open_start()

    def _on_copy_pkgref_to_clip(self) -> None:
        QtWidgets.QApplication.clipboard().setText(self._ui.pbinary_pkgref.text())

    def _on_file_menu(self, position: QtCore.QPoint) -> None:
        selection = self._ui.package_binary.selectedIndexes()
        assert len(selection) == 1
        node = selection[0].internalPointer()
        if not node.is_file:
            return
        if node.link_target:
            return
        menu = QtWidgets.QMenu(self)
        save_action = QtGui.QAction("Save ...", self)
        save_action.triggered.connect(self._on_file_save)
        menu.addAction(save_action)
        menu.exec_(self.sender().mapToGlobal(position))

    def _on_file_save(self) -> None:
        selection = self._ui.package_binary.selectedIndexes()
        assert len(selection) == 1
        node = selection[0].internalPointer()
        new_path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, f"Save {node.path} from package", "", ""
        )
        if not new_path:
            return
        if node.container:
            with tarfile.open(self._artifact_folder / node.container, "r") as tar:
                contents_object = tar.extractfile(node.tar_info)
                assert contents_object
                new_path_path = pathlib.Path(new_path)
                with new_path_path.open("wb") as writer:
                    writer.write(contents_object.read())
        else:
            shutil.copyfile(self._artifact_folder / node.path, new_path)



#!/usr/bin/env python3

"""
Create a lockfile but in memory
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.packagenode import PackageNode
from cruiz.interop.dependencygraph import DependencyGraph
from cruiz.interop.message import Message, Success

from cruiz.workers.utils.formatoptions import format_options_v2

from . import worker


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan lock create'
    """
    with worker.ConanWorker(queue, params) as api:
        try:
            import dataclasses
            import os

            from conans.client.graph.graph import (
                RECIPE_CONSUMER,
                RECIPE_VIRTUAL,
            )

            remotes = api.remotes.list()

            # fake command line arguments
            @dataclasses.dataclass
            class FakeCLIArguments:
                name = params.name
                version = params.version
                user = params.user
                channel = params.channel
                build: None = None
                update = True
                build_require: None = None
                profile_build: None = None
                profile_host: None = None
                settings_build: None = None
                options_build = format_options_v2(params.options)
                conf_build: None = None
                settings_host: None = None
                options_host = format_options_v2(params.options)
                conf_host: None = None

            args = FakeCLIArguments()

            profile_host, profile_build = api.profiles.get_profiles_from_args(args)

            lockfile = None
            # first get enough information about recipes
            assert params.recipe_path
            deps_graph = api.graph.load_graph_consumer(
                os.fspath(params.recipe_path),
                args.name,
                args.version,
                args.user,
                args.channel,
                profile_host,
                profile_build,
                lockfile,
                remotes,
                args.build,
                args.update,
                is_build_require=args.build_require,
            )

            # then get binary information, e.g. package_ids
            api.graph.analyze_binaries(
                deps_graph,
                args.build,
                remotes=remotes,
                update=args.update,
                lockfile=lockfile,
            )

            # create nodes (derived from conan.graph.printer.print_graph)
            # continuing to use this in Conan 2
            build_time_nodes = deps_graph.build_time_nodes()
            nodes = {}
            for node in sorted(deps_graph.nodes):
                info = node.conanfile.original_info.dumps()
                build_folder = node.conanfile.folders.build
                # in Conan 2, there are no short paths
                short_paths = False

                if node.conanfile.info.invalid:
                    raise ValueError(node.conanfile.info.invalid)

                if node.recipe in (RECIPE_CONSUMER, RECIPE_VIRTUAL):
                    new_node = PackageNode(
                        node.name,
                        str(node.ref),
                        node.package_id,
                        node.ref.revision,
                        short_paths,
                        info,
                        True,
                        build_folder,
                    )
                    root_node = nodes[node] = new_node
                    continue
                is_runtime = node not in build_time_nodes
                new_node = PackageNode(
                    node.name,
                    str(node.ref),
                    node.package_id,
                    node.ref.revision,
                    short_paths,
                    info,
                    is_runtime,
                    build_folder,
                )
                nodes[node] = new_node

            # connect nodes (two-way)
            for node, interop_node in nodes.items():
                for dep in node.dependencies:
                    interop_node.children.append(nodes[dep.dst])
                    nodes[dep.dst].parents.append(interop_node)

            new_graph = DependencyGraph(list(nodes.values()), root_node)

            queue.put(Success(new_graph))
        except Exception as exc:
            print(exc)
            raise



#!/usr/bin/env python3

"""
Conan config entries of interest
"""

from enum import Enum


class ConanConfigBoolean(Enum):
    """
    Enumeration for Conan configuration items that are Boolean values

    Note that these appear to be Conan 1 specific

    PRINT_RUN_COMMANDS is always on
    (see https://github.com/conan-io/conan/issues/6752#issuecomment-1570619957)

    REVISIONS are always on
    """

    PRINT_RUN_COMMANDS = "log.print_run_commands"
    REVISIONS = "general.revisions_enabled"



#!/usr/bin/env python3

"""
Search recipes parameters
"""

import typing

from .commonparameters import CommonParameters


class SearchRecipesParameters(CommonParameters):
    """
    Representation of all the arguments to search for a recipe

    Attributes are dynamically added from the keyword-args passed in.
    """

    def __init__(self, **args: typing.Any) -> None:
        import cruiz.workers.api as workers_api

        super().__init__(workers_api.remotesearch.invoke)
        for k, v in args.items():
            self.__setattr__(k, v)



#!/usr/bin/env python3

"""
Get package_ids for a given package reference with revision
"""

from __future__ import annotations

import multiprocessing
import typing

from cruiz.interop.packageidparameters import PackageIdParameters
from cruiz.interop.message import Message, Success

from . import worker


def invoke(queue: multiprocessing.Queue[Message], params: PackageIdParameters) -> None:
    """
    Equivalent to

    'conan search -r <remote_name> <reference>'

    PackageIdParameters has dynamic attributes.
    """
    with worker.ConanWorker(queue, params) as api:
        from conan.internal.conan_app import ConanApp
        from conans.model.recipe_ref import RecipeReference

        remote = api.remotes.get(params.remote_name)
        ref = RecipeReference.loads(params.reference)
        app = ConanApp(api.cache_folder)

        # dict of keys of type <class 'conans.model.package_ref.PkgReference'>
        results_dict = app.remote_manager.search_packages(
            remote=remote,
            ref=ref,
        )

        results_list: typing.List[typing.Dict[str, str]] = []
        for ref_pkgid, value in results_dict.items():
            entry = {"id": ref_pkgid.package_id}
            entry.update(value)
            results_list.append(entry)

        queue.put(Success(results_list or None))



#!/usr/bin/env python3

"""
Worker for runnin conan export-pkg
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.message import Message, Success

from . import worker


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan export-pkg'
    """
    with worker.ConanWorker(queue, params) as api:
        import importlib

        # although there is a Python API for export-pkg, it works via a dependency
        # graph, so use the CLI interface
        imported_module = importlib.import_module("conan.cli.commands.export_pkg")
        # strip the 'verb' from the front of the argument list
        args = params.to_args()[1:]
        imported_module.export_pkg.run(api, args)

        queue.put(Success(None))



#!/usr/bin/env python3

"""
Process requests for searching a remote for package references
"""

from __future__ import annotations

import multiprocessing
import typing

from cruiz.interop.searchrecipesparameters import SearchRecipesParameters
from cruiz.interop.message import Message, Success

from . import worker


def invoke(
    queue: multiprocessing.Queue[Message], params: SearchRecipesParameters
) -> None:
    """
    Equivalent to:
    'conan search -r <remote_name> [--case-sensitive] <pattern>'

    with optional extra processing when aliases are detected.

    SearchRecipesParameters has dynamic attributes.
    """
    with worker.ConanWorker(queue, params) as api:
        remote_name = params.remote_name  # type: ignore
        remote = api.remotes.get(remote_name)
        refs: typing.List[typing.Tuple[str, typing.Optional[str]]] = []
        for ref in api.search.recipes(query=params.pattern, remote=remote):
            # returns a list of conans.model.recipe_ref.RecipeReference
            refs.append((str(ref), None))

        # TODO: alias aware

        queue.put(Success(refs or None))



#!/usr/bin/env python3

"""
Utils for worker context managers for Conan v1
"""

import contextlib
import typing

from cruiz.workers.utils.worker import Worker


# pylint: disable=too-few-public-methods
class ConanWorker(Worker):
    """
    Conan specific context manager
    """

    def __enter__(self) -> typing.Any:
        super().__enter__()
        # import here because it can use the environment variables set in the
        # super class
        # pylint: disable=import-outside-toplevel
        from .conanapi import instance

        api = instance(self._queue, self._params)
        return api


def replace_conan_version_struct_with_string(
    result: typing.Dict[typing.Any, typing.Any]
) -> None:
    """
    Results have ConanVersion structs, but these cannot be passed over the
    process divide so just convert them to strings
    """
    if "installed" not in result:
        return
    for installed in result["installed"]:
        recipe = installed["recipe"]
        recipe["version"] = str(recipe["version"])
        for package in installed["packages"]:
            try:
                cpp_info = package["cpp_info"]
            except KeyError:
                # no cpp_info, so abort further processing
                continue
            cpp_info["version"] = str(cpp_info["version"])
            with contextlib.suppress(KeyError):
                # ignore 'components' not being in cpp_info
                for component_key in cpp_info["components"]:
                    component = cpp_info["components"][component_key]
                    component["version"] = str(component["version"])
            with contextlib.suppress(KeyError):
                # ignore 'build_modules' not being in cpp_info
                build_modules: typing.Dict[str, str] = {}
                for key, value in cpp_info["build_modules"].items():
                    build_modules[key] = value
                cpp_info["build_modules"] = build_modules



#!/usr/bin/env python3

"""
QLineEdit subclass that has a custom context menu
"""

import typing

from qtpy import QtGui, QtWidgets


class LineEditWithCustomContextMenu(QtWidgets.QLineEdit):
    """
    Subclass allowing a custom context menu to be appended to the standard menu.
    """

    def add_submenu_actions(
        self, submenu_name: str, actions: typing.List[QtGui.QAction]
    ) -> None:
        """
        Provide the name of a submenu and the QActions for it
        """
        self._submenu_name = submenu_name
        self._submenu_actions = actions

    def contextMenuEvent(self, event: QtGui.QContextMenuEvent) -> None:
        menu = self.createStandardContextMenu()
        menu.addSeparator()
        submenu = menu.addMenu(self._submenu_name)
        if self._submenu_actions:
            submenu.addActions(self._submenu_actions)
        else:
            submenu.setEnabled(False)

        menu.exec_(event.globalPos())



#!/usr/bin/env python3

"""
Wrapper around a child process, that can process multiple requests for
meta data from a Conan instance
"""

import logging
import multiprocessing
import os
import sys
import typing
import urllib.parse

from qtpy import QtCore

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.message import (
    Success,
    Failure,
    Stdout,
    Stderr,
    ConanLogMessage,
)
import cruiz.workers.api as workers_api
from cruiz.dumpobjecttypes import dump_object_types

from .conanenv import get_conan_env
from .logdetails import LogDetails


logger = logging.getLogger(__name__)


class MetaRequestConanInvocation(QtCore.QObject):
    """
    Wrapper around request-reply interaction for meta data from a Conan local cache
    Request-replies are synchronous, since it's just metadata
    """

    def __del__(self) -> None:
        logger.debug("-=%d", id(self))

    def __init__(  # noqa: F811
        self,
        parent: QtCore.QObject,
        cache_name: str,
        log_details: LogDetails,
    ) -> None:
        logger.debug("+=%d", id(self))
        super().__init__(parent)
        self._log_details = log_details
        self._mp_context = multiprocessing.get_context("spawn")
        self._request_queue = self._mp_context.JoinableQueue()
        self._reply_queue = self._mp_context.Queue()
        self.active = False
        params = CommandParameters("meta", workers_api.meta.invoke)
        added_environment, removed_environment = get_conan_env(cache_name)
        params.added_environment.update(added_environment)
        params.removed_environment.extend(removed_environment)
        self._invoke_conan_process(params)

    def close(self) -> None:
        """
        Close all resources associated with the invocation
        """
        logger.debug("(%d) closing request queue...", id(self))
        self._request_queue.put("end")
        self._request_queue.join()
        self._request_queue.close()
        self._request_queue.join_thread()
        logger.debug("(%d) closing reply queue...", id(self))
        self._reply_queue.close()
        self._reply_queue.join_thread()
        logger.debug("(%d) joining process...", id(self))
        self._process.join()
        self._process.close()

    def _invoke_conan_process(self, params: CommandParameters) -> None:
        process = self._mp_context.Process(
            target=params.worker,
            args=(self._request_queue, self._reply_queue, params),
            daemon=False,
        )
        process.start()
        logger.debug(
            "cruiz (pid=%i) started child process (pid=%i) for %s",
            os.getpid(),
            process.pid,
            params.worker.__module__,
        )
        self._process = process

    @staticmethod
    def __check_for_conan_leakage(entry: typing.Any = None) -> None:
        if "conans" not in sys.modules:
            return
        if entry:
            logger.critical("Conan has leaked into message queue object dump:")
            dump_object_types(entry, loglevel="CRITICAL")
        logger.critical("Conan has leaked into cruiz")
        sys.exit(1)

    def request_data(
        self,
        request: typing.Any,
        params: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ) -> typing.Tuple[typing.Any, typing.Optional[Exception]]:
        """
        Request a named metadata, and synchronously wait for a reply.
        Reply is a tuple, with result and exception.
        """
        meta_request = (
            f"{request}?{urllib.parse.urlencode(params, doseq=True)}"
            if params
            else request
        )
        logger.debug("* Requesting %s", meta_request)
        self._request_queue.put(meta_request)
        self.active = True
        logger.debug("* Requested %s waiting for reply", meta_request)
        # the requested task is done when there is a success or a failure
        self._request_queue.join()
        response: typing.Union[
            Success,
            Failure,
            None,
        ] = None
        while True:
            MetaRequestConanInvocation.__check_for_conan_leakage(None)
            reply = self._reply_queue.get()
            MetaRequestConanInvocation.__check_for_conan_leakage(reply)
            if isinstance(reply, Stdout):
                logger.debug("* Got stdout message: '%s", reply.message())
                self._log_details.stdout(reply.message())
            elif isinstance(reply, Stderr):
                logger.debug("* Got stderr message: '%s", reply.message())
                self._log_details.stderr(reply.message())
            elif isinstance(reply, ConanLogMessage):
                logger.debug("* Got Conan log message: '%s", reply.message())
                self._log_details.conan_log(reply.message())
            elif isinstance(reply, Success):
                logger.debug("* Requested %s got reply '%s'", meta_request, reply)
                response = reply
                break
            elif isinstance(reply, Failure):
                logger.debug("* Got failure message: '%s'", str(reply.exception()))
                response = reply
                break
        assert self._reply_queue.empty()
        self.active = False
        if response is not None:
            if isinstance(reply, Success):
                return (response.payload(), None)  # type: ignore
            if isinstance(reply, Failure):
                return (None, response.exception())  # type: ignore
        raise RuntimeError("No success message")



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.message import Message, Success

from . import worker


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run an arbitrary command
    """
    with worker.ConanWorker(queue, params) as api:
        from conans.client.command import Command

        command = Command(api)
        verb_function = getattr(command, params.verb)
        result = verb_function(params.arguments)
        queue.put(Success(result))



#!/usr/bin/env python3

"""
Get source code for the package
"""

from __future__ import annotations

import multiprocessing
import os

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.message import Message, Success

from . import worker


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan source'
    """
    with worker.ConanWorker(queue, params) as api:
        assert params.recipe_path
        assert params.cwd

        args = {}
        if params.name:
            args["name"] = params.name
        if params.version:
            args["version"] = params.version
        if params.user:
            args["user"] = params.user
        if params.channel:
            args["channel"] = params.channel

        api.local.source(os.fspath(params.recipe_path), **args)
        queue.put(Success(None))



#!/usr/bin/env python3

"""
Get the package binary for a given:
reference, recipe revision, package_id, package revision
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.packagebinaryparameters import PackageBinaryParameters
from cruiz.interop.message import Message, Success

from . import worker


def invoke(
    queue: multiprocessing.Queue[Message], params: PackageBinaryParameters
) -> None:
    """
    Similar to 'conan download' but the download folder is custom.

    PackageBinaryParameters has dynamic attributes.
    """
    # pylint: disable=import-outside-toplevel, protected-access, no-member
    with worker.ConanWorker(queue, params) as api:
        from conan.internal.conan_app import ConanApp
        from conans.model.package_ref import PkgReference

        remote = api.remotes.get(params.remote_name)
        pref = PkgReference.loads(params.reference)
        app = ConanApp(api.cache_folder)
        metadata = None

        # TODO: using non-public method
        zipped_files = app.remote_manager._call_remote(
            remote,
            "get_package",
            pref,
            params.where,
            metadata=metadata,
            only_metadata=False,
        )

        queue.put(Success(zipped_files))



#!/usr/bin/env python3

"""
Global variables
(could do without these)
"""

from __future__ import annotations
import typing

CRUIZ_MAINWINDOW: typing.Optional[cruiz.MainWindow] = None  # type: ignore # noqa: F821

CONAN_MAJOR_VERSION: int = 0

CRUIZ_THEME: typing.Optional[str] = None


def get_main_window() -> cruiz.MainWindow:  # type: ignore # noqa: F821
    """
    Get the main window for the application, asserting it is valid
    """
    assert CRUIZ_MAINWINDOW
    return CRUIZ_MAINWINDOW


def get_theme() -> str:
    """
    Get the name of the current theme
    """
    assert CRUIZ_THEME
    return CRUIZ_THEME


def is_dark_theme() -> bool:
    """
    Is the current theme the dark theme?
    """
    assert CRUIZ_THEME
    return CRUIZ_THEME == "Dark"


def __capture_conan_version() -> None:
    global CONAN_MAJOR_VERSION
    if CONAN_MAJOR_VERSION > 0:
        return

    import subprocess

    get_conan_version = subprocess.run(
        ["conan", "--version"], capture_output=True, encoding="utf8", errors="ignore"
    )
    version = get_conan_version.stdout.replace("Conan version ", "").strip()
    if not version:
        raise ValueError("Unable to determine Conan version")
    version_components = version.split(".")
    CONAN_MAJOR_VERSION = int(version_components[0])


__capture_conan_version()



#!/usr/bin/env python3

"""
Wizard page for selecting the version of the package
"""

import typing

from qtpy import QtGui, QtWidgets

from cruiz.widgets.util import BlockSignals
from cruiz.settings.managers.conanpreferences import ConanSettingsReader
from cruiz.settings.managers.recipe import RecipeSettingsReader

import cruiz.globals


class LoadRecipePackageVersionPage(QtWidgets.QWizardPage):
    """
    Wizard page for selecting the recipe version to bind to.
    """

    @property
    def _ui(self) -> typing.Any:
        return self.wizard().ui

    def nextId(self) -> int:
        if self._ui.version.currentText() in self._uuid_versions:
            return -1
        return 2

    def initializePage(self) -> None:
        self.registerField("version*", self._ui.version, "currentText")  # type: ignore

        self._ui.version.currentTextChanged.connect(self._on_version_changed)

        self._uuid_versions: typing.Dict[str, str] = {}
        self._resolve_ui_to_possible_versions()
        super().initializePage()

    def cleanupPage(self) -> None:
        self._ui.version.currentTextChanged.disconnect()
        return super().cleanupPage()

    def _on_version_changed(self, text: str) -> None:
        if text in self._uuid_versions:
            self.setFinalPage(True)
            self.wizard().uuid = self._uuid_versions[text]
        else:
            self.wizard().uuid = None
        self.completeChanged.emit()

    def _resolve_ui_to_possible_versions(self) -> None:
        self._ui.version.setEditable(False)
        version_in_recipe = self.wizard().recipe_attributes.get("version")
        if version_in_recipe is None:
            # get versions from conandata.yml
            conandata = self.wizard().conandata
            if conandata:
                versions = self._get_versions_from_conandata(conandata)
                if versions:
                    # get the versions from uuids with the same paths
                    for uuid in self.wizard().matching_uuids:
                        with RecipeSettingsReader.from_uuid(uuid) as settings:
                            attributes = settings.attribute_overrides.resolve()
                        if "version" in attributes:
                            self._uuid_versions[attributes["version"]] = uuid

                    with BlockSignals(self._ui.version) as blocked_widget:
                        blocked_widget.clear()
                        model = blocked_widget.model()
                        for i, version in enumerate(versions):
                            if version in self._uuid_versions:
                                blocked_widget.addItem(
                                    QtGui.QIcon(":/cruiz.png"), version
                                )
                                if cruiz.globals.get_main_window().is_recipe_active(
                                    self._uuid_versions[version]
                                ):
                                    item = model.item(i)
                                    item.setFlags(
                                        item.flags() & ~QtGui.Qt.ItemIsEnabled
                                    )
                            else:
                                blocked_widget.addItem(version)
                        blocked_widget.setCurrentIndex(-1)
            else:
                # if the conandata.yml is not available, manually specify a version
                self._ui.version.setEditable(True)
                self._ui.version.lineEdit().setPlaceholderText(
                    "Package version to use, e.g. 1.2.3"
                )
        else:
            with BlockSignals(self._ui.version) as blocked_widget:
                blocked_widget.clear()
                blocked_widget.addItem(version_in_recipe)
            # combobox disabled to ensure it's clear that there's no choice
            self._ui.version.setEnabled(False)

    def _get_versions_from_conandata(
        self, conandata: typing.Dict[str, typing.Dict[str, str]]
    ) -> typing.Optional[typing.List[str]]:
        assert conandata
        with ConanSettingsReader() as settings:
            path_segments = settings.conandata_version_yaml_pathsegment.resolve()
        if not path_segments:
            QtWidgets.QMessageBox.information(
                self,
                "Cannot identify recipe version numbers from conandata.yml file",
                "The conandata.yml beside the recipe cannot be parsed for version "
                "numbers.\n\n"
                "This is because cruiz cannot determine its format without a hint.\n\n"
                "Complete the 'Preferences->Conan->Parsing conandata' setting to "
                "specify how to find the list of version numbers for your recipe.\n\n"
                "For example, if your conandata.yml is formatted as\n\n"
                "\tsources:\n"
                "\t  1.2.3:\n"
                "\t    ...\n"
                "\t  4.5.6:\n"
                "\t    ...\n\n"
                "then specify 'sources' in the preferences.",
            )
            return None
        current_dict = conandata
        for key in path_segments.split("/"):
            if key in current_dict:
                current_dict = conandata[key]  # type: ignore[assignment]
            else:
                QtWidgets.QMessageBox.information(
                    self,
                    "Cannot find YAML path segment in recipe conandata.yml file",
                    f"Unable to locate YAML path segment '{key}' in conandata.yml "
                    f" from preferences expression '{path_segments}'.\n\nThe version "
                    "number can be manually specified, or the preference corrected for "
                    "the YAML layout.",
                )
                return None
        return list(current_dict.keys())



#!/usr/bin/env python3

"""
Meta commands are short lived queries that do not warrant having their own
process spun up to resolve.

One long-lived meta process runs continually to service these.
"""

from __future__ import annotations

import multiprocessing
import os
import pathlib
import urllib.parse
import typing

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.message import Message, Failure, Success
from cruiz.interop.pod import ConanRemote, ConanHook

from . import worker


def _interop_remote_list(api: typing.Any) -> typing.List[ConanRemote]:
    interop_list: typing.List[ConanRemote] = []
    for remote in api.remotes.list(only_enabled=False):
        interop_list.append(ConanRemote(remote.name, remote.url, not remote.disabled))
    return interop_list


def _interop_remotes_sync(api: typing.Any, remotes: typing.List[str]) -> None:
    from conans.client.cache.remote_registry import Remote

    for remote in _interop_remote_list(api):
        api.remotes.remove(remote.name)
    for remote_str in remotes:
        remote = ConanRemote.from_string(remote_str)
        conan_remote = Remote(remote.name, remote.url)
        api.remotes.add(conan_remote)
        if remote.enabled:
            api.remotes.enable(remote.name)
        else:
            api.remotes.disable(remote.name)


def _interop_get_config(api: typing.Any, key: str) -> typing.Optional[str]:
    return api.config.get(key)


def _interop_profiles_dir(api: typing.Any) -> pathlib.Path:
    from conan.internal.conan_app import ConanApp

    app = ConanApp(api.cache_folder)
    return pathlib.Path(app.cache.profiles_path)


def _interop_get_hooks(api: typing.Any) -> typing.List[ConanHook]:
    from conan.internal.conan_app import ConanApp

    app = ConanApp(api.cache_folder)

    hooks_dir = app.cache.hooks_path

    hook_files: typing.List[ConanHook] = []
    for root, dirs, files in os.walk(hooks_dir):
        if ".git" in dirs:
            # ignore .git folders
            dirs.remove(".git")
        for file in files:
            if file.endswith(".py"):
                full_path = pathlib.Path(root) / file
                stored_path = full_path.relative_to(hooks_dir)
                # in Conan 2, hooks are activated by having them in the folder
                hook_files.append(ConanHook(stored_path, enabled=True))

    return hook_files


def _interop_inspect_recipe(
    api: typing.Any, recipe_path: str
) -> typing.Dict[str, typing.Any]:
    conanfile = api.local.inspect(recipe_path, None, None)
    result = conanfile.serialize()
    return result


def _interop_create_default_profile(api: typing.Any) -> None:
    from conans.util.files import save

    profile_pathname = api.profiles.get_path("default", os.getcwd(), exists=False)
    save(profile_pathname, api.profiles.detect().dumps())


def _interop_get_conandata(
    api: typing.Any, recipe_path: str
) -> typing.Dict[str, typing.Any]:
    from conan.internal.conan_app import ConanApp

    app = ConanApp(api.cache_folder)
    return app.loader._load_data(recipe_path)


def _interop_get_config_envvars(api: typing.Any) -> typing.List[str]:
    # no variable present listing the all in Conan 2
    # so use https://docs.conan.io/2/reference/environment.html
    envvar_list = [
        "CONAN_HOME",
        "CONAN_DEFAULT_PROFILE",
        "CONAN_LOGIN_USERNAME",
        "CONAN_PASSWORD",
        "CONAN_COLOR_DARK",
    ]
    return envvar_list


def _interop_profile_meta(
    api: typing.Any, profile: str
) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
    from conans.client.cache.cache import ClientCache
    from conans.client.profile_loader import ProfileLoader

    cache = ClientCache(api.cache_folder)
    loader = ProfileLoader(cache)
    # TODO: using internal method
    profile = loader._load_profile(profile, os.getcwd())
    details: typing.Dict[str, typing.Dict[str, typing.Any]] = {"settings": {}}
    for key, value in profile.settings.items():
        details["settings"][key] = value
    return details


def invoke(
    request_queue: multiprocessing.JoinableQueue[str],
    reply_queue: multiprocessing.Queue[Message],
    params: CommandParameters,
) -> None:
    """
    Run continuous loop, waiting on requests from the main process
    """
    with worker.ConanWorker(reply_queue, params) as api:
        while True:
            try:
                request = request_queue.get()
                if request == "end":
                    request_queue.task_done()
                    break
                if "?" in request:
                    split = request.split("?")
                    request = split[0]
                    request_params = urllib.parse.parse_qs(split[1])

                result: typing.Any = None
                if request == "remotes_list":
                    result = _interop_remote_list(api)
                elif request == "remotes_sync":
                    _interop_remotes_sync(api, request_params["remotes"])
                    result = None
                elif request == "get_config":
                    result = _interop_get_config(api, request_params["config"][0])
                elif request == "profiles_dir":
                    result = _interop_profiles_dir(api)
                elif request == "get_hooks":
                    result = _interop_get_hooks(api)
                elif request == "inspect_recipe":
                    result = _interop_inspect_recipe(api, request_params["path"][0])
                elif request == "create_default_profile":
                    _interop_create_default_profile(api)
                    result = None
                elif request == "get_conandata":
                    result = _interop_get_conandata(api, request_params["path"][0])
                elif request == "get_config_envvars":
                    result = _interop_get_config_envvars(api)
                elif request == "profile_meta":
                    result = _interop_profile_meta(api, request_params["name"][0])
                else:
                    raise RuntimeError(
                        f"Meta command request not implemented: '{request}' "
                        f"with params '{request_params}'"
                    )
                reply_queue.put(Success(result))
                request_queue.task_done()
                # ensure that the result doesn't accidentally appear in
                # subsequent loop iterations
                del result
            # pylint: disable=broad-except
            except Exception as exception:
                reply_queue.put(Failure(Exception(str(exception))))
                request_queue.task_done()
    request_queue.join()
    request_queue.close()
    request_queue.join_thread()



#!/usr/bin/env python3

"""
Utils for worker context managers for Conan v2
"""

from __future__ import annotations

import multiprocessing
import subprocess
import typing

from cruiz.workers.utils.worker import Worker
from cruiz.workers.utils.stream import QueuedStreamSix

from cruiz.interop.message import Message, Stdout, Stderr


def _patch_conan_output_initialiser(queue: multiprocessing.Queue[Message]) -> None:
    # this has to be the first import of ConanOutput
    from conan.api.output import ConanOutput

    conanoutput_old_init = ConanOutput.__init__

    def new_init(self: ConanOutput, scope: str = "") -> None:
        conanoutput_old_init(self, scope)
        self.stream = QueuedStreamSix(queue, Stdout)

    ConanOutput.__init__ = new_init


def _patch_conan_run(queue: multiprocessing.Queue[Message]) -> None:
    # this has to be the first import of conan_run
    import conans

    # entirely replacing the vanilla conan_run, because it uses subprocess communicate
    # which does not stream the output, but waits for the end of the process
    def new_conan_run(  # type: ignore[no-untyped-def]
        command, stdout=None, stderr=None, cwd=None, shell=True
    ):
        with conans.util.runners.pyinstaller_bundle_env_cleaned():
            with subprocess.Popen(
                command,
                shell=shell,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=cwd,
            ) as process:
                assert process.stdout
                for line in iter(process.stdout.readline, b""):
                    queue.put(Stdout(line.decode("utf-8")))
                assert process.stderr
                for line in iter(process.stderr.readline, b""):
                    queue.put(Stderr(line.decode("utf-8")))

            return process.returncode

    conans.util.runners.conan_run = new_conan_run


def _do_patching(queue: multiprocessing.Queue[Message]) -> None:
    _patch_conan_output_initialiser(queue)
    _patch_conan_run(queue)


# pylint: disable=too-few-public-methods
class ConanWorker(Worker):
    """
    Conan specific context manager
    """

    def __enter__(self) -> typing.Any:
        super().__enter__()
        # import here because it can use the environment variables
        # set in the super class
        # pylint: disable=import-outside-toplevel
        from conan.api.conan_api import ConanAPI

        _do_patching(self._queue)

        return ConanAPI()



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.packageidparameters import PackageIdParameters
from cruiz.interop.message import Message, Success

from . import worker


def invoke(queue: multiprocessing.Queue[Message], params: PackageIdParameters) -> None:
    """
    Equivalent to

    'conan search -r <remote_name> <reference>'

    PackageIdParameters has dynamic attributes.
    """
    with worker.ConanWorker(queue, params) as api:
        result = api.search_packages(
            params.reference,  # type: ignore
            remote_name=params.remote_name,  # type: ignore
        )
        results_list = result["results"][0]["items"][0]["packages"]

        import conans

        for entry in results_list:
            if "options" in entry:
                new_options = {}
                for option_key, option_value in entry["options"].items():
                    if isinstance(
                        option_value, conans.model.options.PackageOptionValue
                    ):
                        new_options[option_key] = str(option_value)
                    else:
                        new_options[option_key] = option_value
                entry["options"] = new_options

        queue.put(Success(results_list or None))



#!/usr/bin/env python3

"""
Invoke conan create for package creation
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.message import Message, Success

from . import worker


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan create'
    """
    with worker.ConanWorker(queue, params) as api:
        import importlib

        # no Python API for create, so reach for the CLI method
        imported_module = importlib.import_module("conan.cli.commands.create")
        # strip the 'verb' from the front of the argument list
        args = params.to_args()[1:]
        imported_module.create.run(api, args)

        queue.put(Success(None))



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.message import Message, Success, Stdout

from . import worker


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan remove [-f] *'
    """
    with worker.ConanWorker(queue, params) as api:
        force = True
        result = api.remove("*", force=force)
        queue.put(
            Stdout(
                "Removed all packages from the local cache" " (forced)" if force else ""
            )
        )

        queue.put(Success(result))



#!/usr/bin/env python3

"""
Utilities for generating SVGs from Conan dependency graphs
"""

import os
import typing

import graphviz
from qtpy import QtCore, QtGui, QtSvg, QtWidgets, QtSvgWidgets

from cruiz.settings.managers.graphvizpreferences import GraphVizSettingsReader
from cruiz.environ import EnvironSaver
from cruiz.interop.packagenode import PackageNode
from cruiz.interop.dependencygraph import DependencyGraph


class DependenciesToDigraph:
    """
    Convert a Conan dependency graph into a GraphViz Digraph
    """

    def __init__(
        self, depgraph: DependencyGraph, rankdir: str, flipped_edges: bool = False
    ) -> None:
        graph = graphviz.Digraph(
            comment="Conan package dependency graph",
            format="svg",
            engine="dot",
        )
        graph.attr(bgcolor="lightsteelblue")
        graph.attr(rankdir=rankdir)
        graph.attr("edge", color="red")
        graph.attr("node", color="blue", style="filled", fillcolor="white")
        self._visited: typing.List[PackageNode] = []
        self._edges: typing.List[typing.Tuple[PackageNode, PackageNode]] = []
        self._visit(graph, depgraph.root, flipped_edges)
        self.digraph = graph

    def _visit(
        self, graph: graphviz.Digraph, node: PackageNode, flipped_edges: bool
    ) -> None:
        def _format_ref(ref: str) -> str:
            return ref.replace("@", "@\n")

        if node not in self._visited:
            node_name_and_id = _format_ref(node.reference)
            if node.is_runtime:
                graph.node(node_name_and_id, id=node_name_and_id)
            else:
                graph.node(node_name_and_id, fillcolor="gray", id=node_name_and_id)
            self._visited.append(node)
        for child in node.children:
            self._visit(graph, child, flipped_edges)
            if (child, node) not in self._edges:
                if flipped_edges:
                    graph.edge(
                        _format_ref(node.reference), _format_ref(child.reference)
                    )
                else:
                    graph.edge(
                        _format_ref(child.reference), _format_ref(node.reference)
                    )
                self._edges.append((child, node))


class DigraphToSVG:
    """
    Convert a Graphviz Digraph to an SVG
    """

    def __init__(self, depstodigraph: DependenciesToDigraph) -> None:
        with EnvironSaver():
            with GraphVizSettingsReader() as settings:
                graphviz_bin_dir = settings.bin_directory.resolve()
            if graphviz_bin_dir:
                os.environ["PATH"] = (
                    QtCore.QDir.toNativeSeparators(graphviz_bin_dir)
                    + os.pathsep
                    + os.environ["PATH"]
                )
            exe_path = QtCore.QStandardPaths.findExecutable("dot")
            if not exe_path:
                # if dot cannot be found, abort before trying any graphviz work
                raise FileNotFoundError(
                    "Cannot find 'dot' to generate dependency graph visualisation"
                )
            self.svg = depstodigraph.digraph.pipe()


class _SVGDialog(QtWidgets.QDialog):
    def __init__(self, renderer: QtSvg.QSvgRenderer) -> None:
        super().__init__()
        self.setWindowTitle("Package dependency graph view")
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._context_menu)
        self._renderer = renderer
        item = QtSvgWidgets.QGraphicsSvgItem()
        item.setSharedRenderer(renderer)
        self._scene = QtWidgets.QGraphicsScene()
        self._scene.addItem(item)
        self._view = QtWidgets.QGraphicsView(self._scene)
        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._view)

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        super().showEvent(event)
        self._view.fitInView(self._scene.sceneRect(), QtCore.Qt.KeepAspectRatio)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        self._view.fitInView(self._scene.sceneRect(), QtCore.Qt.KeepAspectRatio)
        super().resizeEvent(event)

    def _context_menu(self, position: QtCore.QPoint) -> None:
        menu = QtWidgets.QMenu(self)
        save_action = QtGui.QAction("Save...", self)
        save_action.triggered.connect(self._on_save)
        menu.addAction(save_action)
        menu.exec_(self.mapToGlobal(position))

    def _on_save(self) -> None:
        new_path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save SVG", "", "SVG files (*.svg)"
        )
        if not new_path:
            return
        generator = QtSvg.QSvgGenerator()
        generator.setFileName(new_path)
        generator.setTitle("Conan package dependency graph")
        generator.setDescription("An SVG diagram of the Conan package dependency graph")
        generator.setSize(self._view.sizeHint())
        painter = QtGui.QPainter()
        if painter.begin(generator):
            self._renderer.render(painter)
            painter.end()


class SVGScene(QtWidgets.QGraphicsScene):
    """
    Wrapper around an SVG scene
    """

    def __init__(self, renderer: QtSvg.QSvgRenderer) -> None:
        super().__init__()
        self._renderer = renderer
        item = QtSvgWidgets.QGraphicsSvgItem()
        item.setSharedRenderer(renderer)
        self.addItem(item)

    def mouseDoubleClickEvent(self, event: QtWidgets.QGraphicsSceneMouseEvent) -> None:
        item = self.itemAt(event.scenePos(), QtGui.QTransform())
        if item and isinstance(item, QtSvgWidgets.QGraphicsSvgItem):
            _SVGDialog(self._renderer).exec_()
            event.setAccepted(True)
        return super().mouseDoubleClickEvent(event)
